// -*- tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*-
// vi: set et ts=4 sw=2 sts=2:

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <dune/common/stdthread.hh>

int main() {
  DUNE_ASSERT_CALL_ONCE();
}
